/**
 * easySchema - easyWSDL toolbox Platform.
 * Copyright (c) 2008,  eBM Websourcing
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package org.ow2.easywsdl.schema.impl;

import java.io.InputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Logger;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.sax.SAXSource;

import org.ow2.easywsdl.schema.api.Schema;
import org.ow2.easywsdl.schema.api.SchemaException;
import org.ow2.easywsdl.schema.api.absItf.AbsItfImport;
import org.ow2.easywsdl.schema.api.absItf.AbsItfInclude;
import org.ow2.easywsdl.schema.api.absItf.AbsItfSchema;
import org.ow2.easywsdl.schema.util.EasyXMLFilter;
import org.ow2.easywsdl.schema.util.StreamHelper;
import org.w3c.dom.Document;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.XMLReaderFactory;

/**
 * @author Nicolas Salatge - eBM WebSourcing
 * @author Gael Blondelle - EBM WebSourcing
 */
public class SchemaReaderImpl implements org.ow2.easywsdl.schema.api.SchemaReader {

	private static Logger log = Logger.getLogger(SchemaReaderImpl.class.getName());


	private SchemaJAXBContext jaxbContext = null;

	private Map<FeatureConstants, Object> features = new HashMap<FeatureConstants, Object>();

	private Map<String, AbsItfSchema> importList = new HashMap<String, AbsItfSchema>();


	/*
	 * Private object initializations
	 */
	public SchemaReaderImpl() throws SchemaException {
		jaxbContext = new SchemaJAXBContext();
		this.features.put(FeatureConstants.VERBOSE, false);
		this.features.put(FeatureConstants.IMPORT_DOCUMENTS, true);
	}



	/**
	 * @return the jaxbContext
	 */
	public JAXBContext getJaxbContext() {
		return this.jaxbContext.getJaxbContext();
	}

	public Schema readSchema(final URI schemaURI) throws SchemaException {
		return this.readSchema(schemaURI, (Map<URI, AbsItfSchema>) null);
	}

	public Schema readSchema(final URI schemaURI, final InputStream input) throws SchemaException {
		return this.readSchema(schemaURI, new InputSource(input), null);
	}

	public Schema readSchema(final URI schemaURI, final InputSource source) throws SchemaException {
		return this.readSchema(schemaURI, source, null);
	}

	public Schema readSchema(final URI schemaURI, Map<URI, AbsItfSchema> imports, boolean deleteImports) throws SchemaException {
		return this.readSchema(schemaURI, new InputSource(StreamHelper.getInputStream(schemaURI)), imports, deleteImports);
	}

	public Schema readSchema(final URI schemaURI, Map<URI, AbsItfSchema> imports) throws SchemaException {
		return this.readSchema(schemaURI, StreamHelper.getInputStream(schemaURI), imports);
	}

	public Schema readSchema(final URI schemaURI, final InputStream input, Map<URI, AbsItfSchema> imports) throws SchemaException {
		return this.readSchema(schemaURI, new InputSource(input), imports);
	}

	public Schema readSchema(final URI schemaURI, final InputSource source, Map<URI, AbsItfSchema> imports) throws SchemaException {
		return this.readSchema(schemaURI, source, imports, true);
	}

	private Schema readSchema(final URI schemaURI, final InputSource source, Map<URI, AbsItfSchema> imports, boolean deleteImports) throws SchemaException {
		Schema desc = null;
		if(deleteImports){
			this.getImportList().clear();
		}
		try {
			log.fine("Loading " + schemaURI);
			XMLReader xmlReader = XMLReaderFactory.createXMLReader();
			EasyXMLFilter filter = new EasyXMLFilter(xmlReader);
			SAXSource saxSource = new SAXSource(filter, source);

			// TODO use SAX validation instead of JAXB validation
			// turn off the JAXB provider's default validation mechanism to
			// avoid duplicate validation
			// SchemaReaderImpl.getUnMarshaller().setValidating( false );

			// unmarshal
			final JAXBElement<org.ow2.easywsdl.schema.org.w3._2001.xmlschema.Schema> schemaBinding = this.jaxbContext.getJaxbContext().createUnmarshaller().unmarshal(saxSource, org.ow2.easywsdl.schema.org.w3._2001.xmlschema.Schema.class);

			final org.ow2.easywsdl.schema.org.w3._2001.xmlschema.Schema def = schemaBinding.getValue();

			desc = new org.ow2.easywsdl.schema.impl.SchemaImpl(schemaURI, def, filter.getNamespaceMapper(), filter.getSchemaLocator(), this.getFeatures(), imports, this);
		} catch (SAXException e) {
			throw new SchemaException("Can not get schema at: " + schemaURI, e);
		} catch (JAXBException e) {
			throw new SchemaException("Can not get schema at: " + schemaURI, e);
		}
		return desc;
	}

	public Schema readSchema(Document doc) throws SchemaException {
		Schema desc = null;
		try {
			URI uri = null;
			if (doc.getDocumentURI() != null) {
				uri = new URI(doc.getDocumentURI());
			} else {
				uri = new URI(".");
			}
			desc = this.readSchema(uri, SAXSource.sourceToInputSource(new DOMSource(doc)));
		} catch (URISyntaxException e) {
			throw new SchemaException(e);
		}
		return desc;
	}

	public Schema readSchema(Document doc, Map<URI, AbsItfSchema> imports) throws SchemaException {
		Schema desc = null;
		try {
			URI uri = null;
			if (doc.getBaseURI() != null) {
				uri = new URI(doc.getBaseURI());
			} else {
				uri = new URI(".");
			}
			desc = this.readSchema(uri, SAXSource.sourceToInputSource(new DOMSource(doc)), imports);
		} catch (URISyntaxException e) {
			throw new SchemaException(e);
		}
		return desc;
	}

	public void setFeature(final FeatureConstants name, final Object value)
	throws IllegalArgumentException {
		this.features.put(name, value);
		SchemaReaderImpl.log.finest("set features: name = " + name + " - value = " + value);
	}

	public Object getFeature(final FeatureConstants name) throws IllegalArgumentException {
		return this.features.get(name);
	}

	public Map<FeatureConstants, Object> getFeatures() {
		return this.features;
	}

	public void setFeatures(final Map<FeatureConstants, Object> features) {
		this.features = features;
	}

	public Map<String, AbsItfSchema> getImportList() {
		return importList;
	}

	public void setImportList(Map<String, AbsItfSchema> importLists) {
		this.importList = importLists;
	}
}
